#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <lua.h>
#include <lauxlib.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/types.h>
#include <stdlib.h>
#include <pwd.h>
#ifdef linux
# include <sys/fsuid.h>
# include <grp.h>
#endif

#include "trap.h"
#include "proc_lua.h"
#include "log.h"
#include "dbbackend.h"

static struct option long_options[] = {
	{"help", no_argument, 0, 'h'},
	{"debug", no_argument, 0, 'd'},
	{"config", required_argument, 0, 'c'},
	{"user", required_argument, 0, 'u'},
	{0, 0, 0, 0}
};

struct conf_options {
	char *db_host;
	char *db_name;
	char *db_user;
	char *db_pass;
	int db_port;
	char *sys_user;
	char *nagios_cmd_path;
	char *pidfile;
} conf = {
	"localhost",
	"trapper",
	"trapper",
	"trapper",
	3306,
	"monitor",
	"/opt/monitor/var/rw/nagios.cmd",
	"/opt/trapper/var/run/processor.run"
};

static int
usage(char *name, int code)
{
	printf("%s - Process SNMP trap messages from netsnmp's mysql backend\n"
		"    -h | --help          Show this message\n"
		"    -d | --debug         Do not fork\n"
		"    -c | --config=<file> Set config file\n",
		name
	);
	exit(code);
	return code;
}


static int
run_handlers(proc_trap *trap)
{
	proc_handler *handlers, *h;
	handlers = get_handlers(trap->oid);
	for (h = handlers; h; h = h->next) {
		proc_result *result;
		int ret = run_handler(trap, h, &result);
		if (~ret & RESULT_IGNORE_RESULT) {
			update_state(result);
			proc_result_free(result);
		}
		if (ret & RESULT_STOP_HANDLING)
			break;
	}
	if (handlers)
		proc_handler_free(handlers);
	return 0;
}

static int
write_pid()
{
	if (!conf.pidfile)
		return 0;

	pid_t pid = getpid();
	char *pidstr;
	if (asprintf(&pidstr, "%d\n", pid) < 0)
		return 1;
	int fd = open(conf.pidfile, O_CREAT|O_WRONLY|O_TRUNC, 0644);
	if (fd != -1) {
		if (write(fd, pidstr, strlen(pidstr)) < 1)
			log_msg(LOG_ERR, "Odd failure creating pidfile");
		close(fd);
	}
	free(pidstr);
	return 0;
}

/* permanently drops all privilege to the given user identity */
static void
drop_privileges(const char *username, uid_t uid)
{
	struct passwd *pwe;

	/*
	 * if we don't have root and can't get it, we can't drop
	 * privs, but we also don't need to, so just return
	 */
	if (setuid(0) != 0)
		return;

	if (!username && !uid) {
		log_msg(LOG_ERR, "No username or uid provided for drop_privileges()");
		abort();
	}
	if (uid)
		pwe = getpwuid(uid);
	else
		pwe = getpwnam(username);

	if (!pwe) {
		log_msg(LOG_ERR, "getpwnam('%s') or getpwuid(%d) failed: %s", username ? username : "(null)", uid, strerror(errno));
		abort();
	}

#ifdef linux
	/* drop ancillary groups first and all uid's last */
	if (setgroups(1, &pwe->pw_gid) < 0) {
		log_msg(LOG_ERR, "setgroups(1, %d) failed: %s", pwe->pw_gid, strerror(errno));
		abort();
	}
	if (setregid(pwe->pw_gid, pwe->pw_gid) < 0) {
		log_msg(LOG_ERR, "setregid(%d, %d) failed: %s", pwe->pw_gid, pwe->pw_gid, strerror(errno));
		abort();
	}
	if (setreuid(pwe->pw_uid, pwe->pw_uid) < 0) {
		log_msg(LOG_ERR, "setreuid(%d, %d) failed: %s", pwe->pw_uid, pwe->pw_uid, strerror(errno));
		abort();
	}
#else
	if (setegid(pwe->pw_gid) < 0) {
		log_msg(LOG_ERR, "setegid(%d) failed: %s", pwe->pw_gid, strerror(errno));
		abort();
	}
	if (setgid(pwe->pw_gid) < 0) {
		log_msg(LOG_ERR, "setgid(%d) failed: %s", pwe->pw_gid, strerror(errno));
		abort();
	}
	if (seteuid(pwe->pw_uid) < 0) {
		log_msg(LOG_ERR, "seteuid(%d) failed: %s", pwe->pw_uid, strerror(errno));
		abort();
	}
	if (setuid(pwe->pw_uid) < 0) {
		log_msg(LOG_ERR, "setuid(%d) failed: %s", pwe->pw_uid, strerror(errno));
		abort();
	}
#endif

	/*
	 * if we can restore root privileges we have either a bug
	 * or some linux caps working against us. In either case,
	 * just abort, but erase user data we hold in memory first.
	 */
	if (!setuid(0) || !seteuid(0)) {
		if (pwe->pw_name)
			memset(pwe->pw_name, 0, strlen(pwe->pw_name));
		if (pwe->pw_passwd)
			memset(pwe->pw_passwd, 0, strlen(pwe->pw_passwd));
		if (pwe->pw_gecos)
			memset(pwe->pw_gecos, 0, strlen(pwe->pw_gecos));
		if (pwe->pw_dir)
			memset(pwe->pw_dir, 0, strlen(pwe->pw_dir));
		if (pwe->pw_shell)
			memset(pwe->pw_shell, 0, strlen(pwe->pw_shell));
		memset(pwe, 0, sizeof(*pwe));
		uid = 0;
		log_msg(LOG_ERR, "Managed to restore root privileges after dropping them. Aborting");
		abort();
	}
}

static void
daemonize()
{
	drop_privileges(conf.sys_user, 0);

	switch (fork()) {
		case -1:
			log_msg(LOG_ERR, "Couldn't fork - exiting");
			exit(1);
			break;
		case 0:
			break;
		default:
			exit(0);
			break;
	}
	if (setsid()  < 0) {
		exit(1);
	}
	switch (fork()) {
		case -1:
			log_msg(LOG_ERR, "Couldn't fork - exiting");
			exit(1);
			break;
		case 0:
			write_pid();
			break;
		default:
			exit(0);
			break;
	}
	umask(0);
	if (chdir("/") < 0) {
		/* i really don't care */
	}
	close(0);
	close(1);
	close(2);
	if (open("/dev/null", O_RDWR) < 0) {
		log_msg(LOG_ERR, "Unable to open /dev/null: %s", strerror(errno));
	}
	if (dup(0) < 0) {
		/* what to do? */
	}
	if (dup(0) < 0) {
		/* what to do? */
	}
}

static int run = 1;


static void
proc_sighandler(int signum __attribute__((unused)))
{
	run = 0;
}

static int
mainloop(dbi_conn conn)
{
	proc_trap *trap;
	int errcode = 0;
	const char *errmsg_dest = NULL;
	while (run) {
		trap = get_traps(conn, &errcode, &errmsg_dest);
		if (trap) {
			while (trap) {
				proc_trap *t = trap;
				char *str = display_trap(t);
				log_msg(LOG_DEBUG, "\n\nFound trap %s", str);
				free(str);
				run_handlers(trap);
				trap = trap->next;
				delete_trap(conn, t);
				proc_trap_free(t);
			}
		} else	if (2006 == errcode) {// 2006: CR_SERVER_GONE_ERROR
			log_msg(LOG_ERR, "  errormsg %s",  errmsg_dest);
			log_msg(LOG_ERR, "  waiting for mysql server to restart ...");
			//reconnect to mysql db till it's up.
			db_disconnect();
			do {
				conn = db_connect(conf.db_host, conf.db_name, conf.db_user, conf.db_pass, conf.db_port);
				if (!conn) {
					log_msg(LOG_ERR, "Could not connect to database, would reconnect after 3 seconds");
					sleep(3);
				}
			} while (!conn);
			log_msg(LOG_INFO, "Connection successfull.");
		} else {
			sleep(1);
		}
		errcode = 0;
	}
	return 0;
}

struct conf_options*
parse_config(char *path)
{
	int res;
	lua_State *L = luaL_newstate();
	if ((res = luaL_loadfile(L, path))) {
		switch (res) {
			case LUA_ERRFILE:
				log_msg(LOG_ERR, "Can not access/read config file %s, exiting.", path);
				break;
			case LUA_ERRSYNTAX:
				log_msg(LOG_ERR, "Can not parse config file %s, exiting.", path);
				break;
			case LUA_ERRMEM:
				log_msg(LOG_ERR, "Memory error parsing config file %s, exiting.", path);
				break;
			default:
				log_msg(LOG_ERR, "Unknown error parsing config file %s, exiting.", path);
				break;
		}
		exit(1);
	}
	lua_newtable(L);
	lua_newtable(L);
	lua_setfield(L, 2, "db");

	lua_newtable(L);
	lua_newtable(L);
	lua_setfield(L, 3, "nagios");
	lua_newtable(L);
	lua_setfield(L, 3, "merlindb"); // this key was present in earlier versions
	lua_setfield(L, 2, "output");
	lua_newtable(L);
	lua_setfield(L, 2, "general"); // as above, unused
	lua_newtable(L);
	lua_setfield(L, 2, "system");
	lua_newtable(L);
	lua_setfield(L, 3, "user");
	lua_setglobal(L, "env");
	lua_getglobal(L, "env");
	lua_setfenv(L, 1);
	if ((res = lua_pcall(L, 0, LUA_MULTRET, 0))) {
		switch (res) {
			case LUA_ERRRUN:
				log_msg(LOG_ERR, "Runtime error in config file %s, exiting.", path);
				break;
			case LUA_ERRERR:
				log_msg(LOG_ERR, "Misc error in error handler parsing config file %s, exiting. Things are probably bonkers somewhere.", path);
				break;
			case LUA_ERRMEM:
				log_msg(LOG_ERR, "Memory error running config file %s, exiting.", path);
				break;
			default:
				log_msg(LOG_ERR, "Unknown error running config file %s, exiting.", path);
				break;
		}
		exit(1);
	}
	lua_getglobal(L, "env");
	lua_getfield(L, 1, "db");
	if (lua_istable(L, lua_gettop(L))) {
		lua_getfield(L, lua_gettop(L), "host");
		if (lua_isstring(L, lua_gettop(L))) {
			conf.db_host = strdup(lua_tostring(L, lua_gettop(L)));
		}
		lua_pop(L, 1);
		lua_getfield(L, lua_gettop(L), "name");
		if (lua_isstring(L, lua_gettop(L))) {
			conf.db_name = strdup(lua_tostring(L, lua_gettop(L)));
		}
		lua_pop(L, 1);
		lua_getfield(L, lua_gettop(L), "user");
		if (lua_isstring(L, lua_gettop(L))) {
			conf.db_user = strdup(lua_tostring(L, lua_gettop(L)));
		}
		lua_pop(L, 1);
		lua_getfield(L, lua_gettop(L), "password");
		if (lua_isstring(L, lua_gettop(L))) {
			conf.db_pass = strdup(lua_tostring(L, lua_gettop(L)));
		}
		lua_pop(L, 1);
		lua_getfield(L, lua_gettop(L), "port");
		if (lua_isnumber(L, lua_gettop(L))) {
			conf.db_port = lua_tonumber(L, lua_gettop(L));
		}
		lua_pop(L, 1);
	}
	lua_pop(L, 1);
	lua_getfield(L, 1, "output");
	if (lua_istable(L, lua_gettop(L))) {
		lua_getfield(L, lua_gettop(L), "nagios");
		if (lua_istable(L, lua_gettop(L))) {
			lua_getfield(L, lua_gettop(L), "cmdfile");
			if (lua_isstring(L, lua_gettop(L))) {
				conf.nagios_cmd_path = strdup(lua_tostring(L, lua_gettop(L)));
			}
			lua_pop(L, 1);
		}
		lua_pop(L, 1);
	}
	lua_pop(L, 1);
	lua_getfield(L, 1, "pidfile");
	if (lua_isstring(L, lua_gettop(L))) {
		conf.pidfile = strdup(lua_tostring(L, lua_gettop(L)));
	}
	lua_getfield(L, 1, "system");
	if (lua_istable(L, lua_gettop(L))) {
		lua_getfield(L, lua_gettop(L), "user");
		if (lua_isstring(L, lua_gettop(L))) {
			conf.sys_user = strdup(lua_tostring(L, lua_gettop(L)));
		}
		lua_pop(L, 1);
	}
	lua_pop(L, 1);
	lua_close(L);
	return &conf;
}

int
main(int argc, char **argv)
{
	int c, option_index = 0;
	int dofork = 1;
	while ((c = getopt_long(argc, argv, "hdc:", long_options, &option_index)) != -1) {
		switch (c) {
			case 0:
				break;
			case 'h':
				return usage(argv[0], 0);
				break;
			case 'c':
				parse_config(optarg);
				break;
			case 'd':
				dofork = 0;
				break;
			case '?':
				break;
			default:
				printf("%d\n", c);
				return usage(argv[0], 1);
				break;
		}
	}

	if (conf.pidfile) {
		pid_t fd = open(conf.pidfile, O_RDONLY);
		if (fd >= 0) {
			char pidstr[16];
			if (read(fd, pidstr, 15) >= 0) {
				int pid = atoi(pidstr);
				if (pid > 0 && kill(pid, 0) >= 0) {
					printf("Processor is already running, shutting down.\n");
					exit(1);
				}
			}
		}
	}

	nagios_connect(conf.nagios_cmd_path);
	dbi_conn conn = db_connect(conf.db_host, conf.db_name, conf.db_user, conf.db_pass, conf.db_port);
	if (!conn) {
		log_msg(LOG_ERR, "Couldn't connect to database, shutting down");
		db_disconnect();
		return 1;
	}

	if (dofork) {
		daemonize();
	} else {
		write_pid();
	}

	signal(SIGINT, proc_sighandler);
	signal(SIGTERM, proc_sighandler);

	/* only log error messages to mysql, log all info (including handler log()s) to syslog */
	open_log(conn, LOG_INFO, LOG_ERR, dofork ? 0 : LOG_DEBUG);
	mainloop(conn);
	db_disconnect();
	unlink(conf.pidfile);
	return 0;
}
